#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import csv
import os
import hashlib
import json
import struct
import subprocess
import sys

from .GlobalConfig import GlobalConfig, InputEndian


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def printQuietless(*args, **kwargs):
    if not GlobalConfig.QUIET:
        print(*args, **kwargs)

def epprintQuietless(*args, **kwargs):
    if not GlobalConfig.QUIET:
        print(*args, file=sys.stderr, **kwargs)


def printVerbose(*args, **kwargs):
    if not GlobalConfig.QUIET and GlobalConfig.VERBOSE:
        print(*args, **kwargs)

def eprintVerbose(*args, **kwargs):
    if not GlobalConfig.QUIET and GlobalConfig.VERBOSE:
        print(*args, file=sys.stderr, **kwargs)

# https://stackoverflow.com/questions/1512457/determining-if-stdout-for-a-python-process-is-redirected
def isStdoutRedirected() -> bool:
    return not sys.stdout.isatty()

# Returns the md5 hash of a bytearray
def getStrHash(byte_array: bytearray) -> str:
    return str(hashlib.md5(byte_array).hexdigest())

def writeBytearrayToFile(filepath: str, array_of_bytes: bytearray):
    with open(filepath, mode="wb") as f:
        f.write(array_of_bytes)

def readFileAsBytearray(filepath: str) -> bytearray:
    if not os.path.exists(filepath):
        return bytearray(0)
    with open(filepath, mode="rb") as f:
        return bytearray(f.read())

def readFile(filepath: str) -> list[str]:
    with open(filepath) as f:
        return [x.strip() for x in f.readlines()]

def readJson(filepath):
    with open(filepath) as f:
        return json.load(f)

def removeExtraWhitespace(line: str) -> str:
    return " ".join(line.split())

def bytesToBEWords(array_of_bytes: bytearray) -> list[int]:
    if GlobalConfig.ENDIAN == InputEndian.MIDDLE:
        # Convert middle endian to big endian
        halfwords = str(int(len(array_of_bytes)//2))
        little_byte_format = f"<{halfwords}H"
        big_byte_format = f">{halfwords}H"
        tmp = struct.unpack_from(little_byte_format, array_of_bytes, 0)
        struct.pack_into(big_byte_format, array_of_bytes, 0, *tmp)

    words = len(array_of_bytes)//4
    endian_format = f">{words}I"
    if GlobalConfig.ENDIAN == InputEndian.LITTLE:
        endian_format = f"<{words}I"
    return list(struct.unpack_from(endian_format, array_of_bytes, 0))

def beWordsToBytes(words_list: list[int], buffer: bytearray) -> bytearray:
    words = len(words_list)
    big_endian_format = f">{words}I"
    struct.pack_into(big_endian_format, buffer, 0, *words_list)
    return buffer

def wordToFloat(word: int) -> float:
    return struct.unpack('>f', struct.pack('>I', word))[0]

def qwordToDouble(qword: int) -> float:
    return struct.unpack('>d', struct.pack('>Q', qword))[0]

def beWordToCurrenEndian(word: int) -> int:
    if GlobalConfig.ENDIAN == InputEndian.BIG:
        return word

    if GlobalConfig.ENDIAN == InputEndian.LITTLE:
        return struct.unpack('<I', struct.pack('>I', word))[0]

    # MIDDLE
    first, second = struct.unpack('>2H', struct.pack('<2H', word >> 16, word & 0xFFFF))
    return (first << 16) | second

def runCommandGetOutput(command: str, args: list[str]) -> list[str] | None:
    try:
        output = subprocess.check_output([command, *args]).decode("utf-8")
        return output.strip().split("\n")
    except:
        return None

def from2Complement(number: int, bits: int) -> int:
    isNegative = number & (1 << (bits - 1))
    if isNegative:
        return -((~number + 1) & ((1 << bits) - 1))
    return number

def readCsv(filepath: str) -> list[list[str]]:
    data: list[list[str]] = []
    with open(filepath) as f:
        lines = f.readlines()
        processedLines = [x.strip().split("#")[0] for x in lines]
        csvReader = csv.reader(processedLines)
        for row in csvReader:
            data.append(list(row))

    return data

def decodeString(buf: bytearray, offset: int) -> tuple[str, int]:
    # Escape characters that are unlikely to be used
    bannedEscapeCharacters = [
        0x01,
        0x02,
        0x03,
        0x04,
        0x05,
        0x06,
        # 0x07, # '\a'
        0x08, # '\b'
        # 0x09, # '\t'
        # 0x0A, # '\n'
        0x0B, # '\v'
        # 0x0C, # '\f'
        # 0x0D, # '\r'
        0x0E,
        0x0F,
        0x10,
        0x11,
        0x12,
        0x13,
        0x14,
        0x15,
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        # 0x1B, # VT escape sequences
        0x1C,
        0x1D,
        0x1E,
        0x1F,
    ]
    dst = bytearray()
    i = 0
    while offset + i < len(buf) and buf[offset + i] != 0:
        dst.append(buf[offset + i])
        i += 1
    if offset + i > len(buf):
        # We reached the end of the buffer without reaching a 0.
        raise RuntimeError()

    for bannedChar in bannedEscapeCharacters:
        if bannedChar in dst:
            raise RuntimeError()

    result = dst.decode("EUC-JP").replace("\n", "\\n").replace("\r", "\\r").replace("\t", "\\t").replace('"', '\\"').replace("\f", "\\f").replace("\a", "\\a").replace("\x1B", "\\x1B")
    return result, i



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from typing import TextIO

from .. import common

from . import symbols


class FileBase(common.ElementBase):
    def __init__(self, context: common.Context, vram: int|None, filename: str, array_of_bytes: bytearray, sectionType: common.FileSectionType):
        super().__init__(context, 0, vram, filename, common.Utils.bytesToBEWords(array_of_bytes), sectionType)

        self.symbolList: list[symbols.SymbolBase] = []

        self.pointersOffsets: set[int] = set()

        self.isHandwritten: bool = False
        self.isRsp: bool = False


    def setCommentOffset(self, commentOffset: int):
        self.commentOffset = commentOffset
        for sym in self.symbolList:
            sym.setCommentOffset(self.commentOffset)

    def getVramOffset(self, localOffset: int) -> int:
        if self.vram is None:
            return self.inFileOffset + localOffset
        # return self.vram + localOffset
        return self.vram + self.inFileOffset + localOffset

    def getAsmPrelude(self) -> str:
        output = ""

        output += ".include \"macro.inc\"" + common.GlobalConfig.LINE_ENDS
        output += common.GlobalConfig.LINE_ENDS
        output += "# assembler directives" + common.GlobalConfig.LINE_ENDS
        output += ".set noat      # allow manual use of $at" + common.GlobalConfig.LINE_ENDS
        output += ".set noreorder # don't insert nops after branches" + common.GlobalConfig.LINE_ENDS
        output += ".set gp=64     # allow use of 64-bit general purpose registers" + common.GlobalConfig.LINE_ENDS
        output += common.GlobalConfig.LINE_ENDS
        output += f".section {self.sectionType.toSectionName()}" + common.GlobalConfig.LINE_ENDS
        output += common.GlobalConfig.LINE_ENDS
        output += ".balign 16" + common.GlobalConfig.LINE_ENDS

        return output

    def getHash(self) -> str:
        buffer = bytearray(4*len(self.words))
        common.Utils.beWordsToBytes(self.words, buffer)
        return common.Utils.getStrHash(buffer)


    def checkAndCreateFirstSymbol(self) -> None:
        "Check if the very start of the file has a symbol and create it if it doesn't exist yet"

        if not common.GlobalConfig.ADD_NEW_SYMBOLS:
            return

        currentVram = self.getVramOffset(0)
        contextSym = self.context.getAnySymbol(currentVram)
        if contextSym is None:
            contextSym = self.context.addSymbol(currentVram, None, self.sectionType, isAutogenerated=True)
            contextSym.isDefined = True


    def printAnalyzisResults(self):
        pass

    def compareToFile(self, other_file: FileBase) -> dict:
        hash_one = self.getHash()
        hash_two = other_file.getHash()

        result = {
            "equal": hash_one == hash_two,
            "hash_one": hash_one,
            "hash_two": hash_two,
            "size_one": self.sizew * 4,
            "size_two": other_file.sizew * 4,
            "diff_bytes": 0,
            "diff_words": 0,
        }

        diff_bytes = 0
        diff_words = 0

        if not result["equal"]:
            min_len = min(self.sizew, other_file.sizew)
            for i in range(min_len):
                for j in range(4):
                    if (self.words[i] & (0xFF << (j * 8))) != (other_file.words[i] & (0xFF << (j * 8))):
                        diff_bytes += 1

            min_len = min(self.sizew, other_file.sizew)
            for i in range(min_len):
                if self.words[i] != other_file.words[i]:
                    diff_words += 1

        result["diff_bytes"] = diff_bytes
        result["diff_words"] = diff_words

        return result

    def blankOutDifferences(self, other: FileBase) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        return False

    def removePointers(self) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        return False


    def disassemble(self) -> str:
        output = ""
        for i, sym in enumerate(self.symbolList):
            output += sym.disassemble()
            if i + 1 < len(self.symbolList):
                output += common.GlobalConfig.LINE_ENDS
        return output

    def disassembleToFile(self, f: TextIO):
        f.write(self.getAsmPrelude())
        f.write(common.GlobalConfig.LINE_ENDS)
        f.write(self.disassemble())


    def saveToFile(self, filepath: str):
        if len(self.symbolList) == 0:
            return

        if filepath == "-":
            self.disassembleToFile(sys.stdout)
        else:
            if common.GlobalConfig.WRITE_BINARY:
                if self.sizew > 0:
                    buffer = bytearray(4*len(self.words))
                    common.Utils.beWordsToBytes(self.words, buffer)
                    common.Utils.writeBytearrayToFile(filepath + self.sectionType.toStr(), buffer)
            with open(filepath + self.sectionType.toStr() + ".s", "w") as f:
                self.disassembleToFile(f)


def createEmptyFile() -> FileBase:
    return FileBase(common.Context(), None, "", bytearray(), common.FileSectionType.Unknown)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import enum


class RelocTypes(enum.Enum):
    INVALID      = -1

    R_MIPS_32    = 2
    R_MIPS_26    = 4
    R_MIPS_HI16  = 5
    R_MIPS_LO16  = 6


    @staticmethod
    def fromValue(value: int) -> RelocTypes:
        if value == 2:
            return RelocTypes.R_MIPS_32
        if value == 4:
            return RelocTypes.R_MIPS_26
        if value == 5:
            return RelocTypes.R_MIPS_HI16
        if value == 6:
            return RelocTypes.R_MIPS_LO16
        return RelocTypes.INVALID



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from . import SymbolBase


class SymbolBss(SymbolBase):
    def __init__(self, context: common.Context, inFileOffset: int, vram: int|None, name: str, spaceSize: int):
        super().__init__(context, inFileOffset, vram, name, list(), common.FileSectionType.Bss)

        self.spaceSize: int = spaceSize


    def disassembleAsBss(self) -> str:
        output = self.getLabel()
        output += self.generateAsmLineComment(0)
        output += f" .space 0x{self.spaceSize:02X}" + common.GlobalConfig.LINE_ENDS
        return output

    def disassemble(self) -> str:
        return self.disassembleAsBss()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
from typing import TextIO

from .. import common

from . import sections
from . import symbols


def createSectionFromSplitEntry(splitEntry: common.FileSplitEntry, array_of_bytes: bytearray, outputPath: str, context: common.Context) -> sections.SectionBase:
    head, tail = os.path.split(outputPath)

    offsetStart = splitEntry.offset
    offsetEnd = splitEntry.nextOffset

    if offsetStart >= 0 and offsetEnd >= 0:
        common.Utils.printVerbose(f"Parsing offset range [{offsetStart:02X}, {offsetEnd:02X}]")
        array_of_bytes = array_of_bytes[offsetStart:offsetEnd]
    elif offsetEnd >= 0:
        common.Utils.printVerbose(f"Parsing until offset 0x{offsetEnd:02X}")
        array_of_bytes = array_of_bytes[:offsetEnd]
    elif offsetStart >= 0:
        common.Utils.printVerbose(f"Parsing since offset 0x{offsetStart:02X}")
        array_of_bytes = array_of_bytes[offsetStart:]

    vram = None
    if splitEntry.vram is not None:
        common.Utils.printVerbose(f"Using VRAM {splitEntry.vram:08X}")
        vram = splitEntry.vram

    f: sections.SectionBase
    if splitEntry.section == common.FileSectionType.Text:
        f = sections.SectionText(context, vram, tail, array_of_bytes)
    elif splitEntry.section == common.FileSectionType.Data:
        f = sections.SectionData(context, vram, tail, array_of_bytes)
    elif splitEntry.section == common.FileSectionType.Rodata:
        f = sections.SectionRodata(context, vram, tail, array_of_bytes)
    elif splitEntry.section == common.FileSectionType.Bss:
        assert isinstance(splitEntry.vram, int)
        f = sections.SectionBss(context, splitEntry.vram, splitEntry.vram + offsetEnd - offsetStart, tail)
    else:
        common.Utils.eprint("Error! Section not set!")
        exit(-1)

    f.isHandwritten = splitEntry.isHandwritten
    f.isRsp = splitEntry.isRsp

    return f

def analyzeSectionFromSplitEntry(fileSection: sections.SectionBase, splitEntry: common.FileSplitEntry):
    offsetStart = splitEntry.offset

    common.Utils.printVerbose("Analyzing")
    fileSection.analyze()
    fileSection.setCommentOffset(offsetStart)

    common.Utils.printVerbose()

    fileSection.printAnalyzisResults()

    return fileSection


def writeSection(path: str, fileSection: sections.SectionBase):
    head, tail = os.path.split(path)

    # Create directories
    if head != "":
        os.makedirs(head, exist_ok=True)

    fileSection.saveToFile(path)

    return path


def getRdataAndLateRodataForFunction(func: symbols.SymbolFunction, rodataFileList: list[sections.SectionRodata]):
    rdataList: list[symbols.SymbolBase] = []
    lateRodataList: list[symbols.SymbolBase] = []
    lateRodataSize = 0

    for rodataSection in rodataFileList:
        if len(rdataList) > 0 or len(lateRodataList) > 0:
            # We already have the rodata for this function. Stop searching
            break

        # Skip the file if there's nothing in this file refenced by the current function
        intersection = func.referencedVRams & rodataSection.symbolsVRams
        if len(intersection) == 0:
            continue

        for rodataSym in rodataSection.symbolList:
            assert rodataSym.vram is not None

            if rodataSym.vram not in intersection:
                continue

            assert rodataSym.contextSym is not None
            # We only care for rodata that's used once
            if rodataSym.contextSym.referenceCounter != 1:
                break

            # A const variable should not be placed with a function
            if rodataSym.contextSym.isMaybeConstVariable():
                break

            if rodataSym.contextSym.isLateRodata():
                lateRodataList.append(rodataSym)
                lateRodataSize += rodataSym.sizew
            else:
                rdataList.append(rodataSym)

    return rdataList, lateRodataList, lateRodataSize

def writeSplittedFunctionToFile(f: TextIO, func: symbols.SymbolFunction, rodataFileList: list[sections.SectionRodata]):
    rdataList, lateRodataList, lateRodataSize = getRdataAndLateRodataForFunction(func, rodataFileList)

    if len(rdataList) > 0:
        # Write the rdata
        f.write(".rdata" + common.GlobalConfig.LINE_ENDS)
        for sym in rdataList:
            f.write(sym.disassemble())
            f.write(common.GlobalConfig.LINE_ENDS)

    if len(lateRodataList) > 0:
        # Write the late_rodata
        f.write(".late_rodata" + common.GlobalConfig.LINE_ENDS)
        if lateRodataSize / len(func.instructions) > 1/3:
            align = 4
            firstLateRodataVram = lateRodataList[0].vram
            if firstLateRodataVram is not None and firstLateRodataVram % 8 == 0:
                align = 8
            f.write(f".late_rodata_alignment {align}" + common.GlobalConfig.LINE_ENDS)
        for sym in lateRodataList:
            f.write(sym.disassemble())
            f.write(common.GlobalConfig.LINE_ENDS)

    if len(rdataList) > 0 or len(lateRodataList) > 0:
        f.write(common.GlobalConfig.LINE_ENDS + ".text" + common.GlobalConfig.LINE_ENDS)

    # Write the function
    f.write(func.disassemble())

def writeSplitedFunction(path: str, func: symbols.SymbolFunction, rodataFileList: list[sections.SectionRodata]):
    os.makedirs(path, exist_ok=True)
    with open(os.path.join(path, func.name) + ".s", "w") as f:
        writeSplittedFunctionToFile(f, func, rodataFileList)


def writeOtherRodata(path: str, rodataFileList: list[sections.SectionRodata]):
    for rodataSection in rodataFileList:
        rodataPath = os.path.join(path, rodataSection.name)
        os.makedirs(rodataPath, exist_ok=True)

        for rodataSym in rodataSection.symbolList:
            if not rodataSym.isRdata():
                continue

            rodataSymbolPath = os.path.join(rodataPath, rodataSym.name) + ".s"
            with open(rodataSymbolPath, "w") as f:
                f.write(".rdata" + common.GlobalConfig.LINE_ENDS)
                f.write(rodataSym.disassemble())



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .GlobalConfig import GlobalConfig
from .Context import Context, ContextSymbolBase
from .FileSectionType import FileSectionType


class ElementBase:
    """Represents the base class used for most file sections and symbols.
    """

    def __init__(self, context: Context, inFileOffset: int, vram: int|None, name: str, words: list[int], sectionType: FileSectionType):
        """Constructor

        Args:
            context (Context):
            inFileOffset (int): The offset of this element relative to the start of its file. It is also used to generate the first column of the disassembled line comment.
            vram (int | None): The VRAM address of this element or `None` if the VRAM is unknown.
            name (str): The name of this element.
            words (list[int]): A list of words (4 bytes) corresponding to this element.
            sectionType (FileSectionType): The section type this element corresponds to.
        """

        self.context: Context = context
        self.inFileOffset: int = inFileOffset
        self.vram: int|None = vram
        self.name: str = name
        self.words: list[int] = words
        self.sectionType: FileSectionType = sectionType

        self.commentOffset: int = 0
        "This value is added to the first column of the disassembled line comment, allowing to change this value without messing inFileOffset"

        self.index: int|None = None
        "The index of the current element inside its parent or `None` if the index is unknown"

        self.parent: ElementBase|None = None
        "For elements that are contained in other elements, like symbols inside of sections"


    @property
    def sizew(self) -> int:
        "The amount of words this element has"
        return len(self.words)

    @property
    def vramEnd(self) -> int|None:
        "The end of this element's VRAM or `None` if VRAM was not specified"
        if self.vram is None:
            return None
        return self.vram + self.sizew * 4


    def setVram(self, vram: int):
        self.vram = vram

    def setCommentOffset(self, commentOffset: int):
        self.commentOffset = commentOffset

    def getVramOffset(self, localOffset: int) -> int:
        if self.vram is None:
            return self.inFileOffset + localOffset
        return self.vram + localOffset
        # return self.vram + self.inFileOffset + localOffset


    def getLabelFromSymbol(self, sym: ContextSymbolBase|None) -> str:
        "Generates a glabel for the passed symbol, including an optional index value if it was set and it is enabled in the GlobalConfig"
        if sym is not None:
            label = sym.getSymbolLabel()
            if GlobalConfig.GLABEL_ASM_COUNT:
                if self.index is not None:
                    label += f" # {self.index}"
            label +=  GlobalConfig.LINE_ENDS
            return label
        return ""


    def analyze(self):
        """Scans the words of this element, gathering as much info as possible.

        This method should be called only once for each element.
        """
        pass


    def disassemble(self) -> str:
        """Produces a disassembly of this element.

        Elements assume the `analyze` method was already called at this point.

        This method can be called as many times as the user wants to.
        """
        return ""



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import dataclasses
import enum


@enum.unique
class InstructionId(enum.Enum):
    INVALID   = -1

    SLL       = enum.auto() # Shift word Left Logical

    SRL       = enum.auto() # Shift word Right Logical
    SRA       = enum.auto() # Shift word Right Arithmetic
    SLLV      = enum.auto() # Shift word Left Logical Variable

    SRLV      = enum.auto() # Shift word Right Logical Variable
    SRAV      = enum.auto() # Shift word Right Arithmetic Variable

    JR        = enum.auto() # Jump Register
    JALR      = enum.auto() # Jump And Link Register
    JALR_RD   = enum.auto() # Jump And Link Register # Special case for rd != 31
    MOVZ      = enum.auto() # MOVe conditional on Zero
    MOVN      = enum.auto() # MOVe conditional on Not zero
    SYSCALL   = enum.auto() # SYStem CALL
    BREAK     = enum.auto() # Break

    SYNC      = enum.auto() # Sync

    MFHI      = enum.auto() # Move From HI register
    MTHI      = enum.auto() # Move To HI register
    MFLO      = enum.auto() # Move From LO register
    MTLO      = enum.auto() # Move To LO register
    DSLLV     = enum.auto() # Doubleword Shift Left Logical Variable

    DSRLV     = enum.auto() # Doubleword Shift Right Logical Variable
    DSRAV     = enum.auto() # Doubleword Shift Right Arithmetic Variable

    MULT      = enum.auto() # MULTtiply word
    MULTU     = enum.auto() # MULTtiply Unsigned word
    DIV       = enum.auto() # DIVide word
    DIVU      = enum.auto() # DIVide Unsigned word
    DMULT     = enum.auto() # Doubleword MULTiply
    DMULTU    = enum.auto() # Doubleword MULTiply Unsigned
    DDIV      = enum.auto() # Doubleword DIVide
    DDIVU     = enum.auto() # Doubleword DIVide Unsigned

    SN64_DIV  = enum.auto() # DIVide word
    SN64_DIVU = enum.auto() # DIVide Unsigned word

    ADD       = enum.auto() # ADD word
    ADDU      = enum.auto() # ADD Unsigned word
    SUB       = enum.auto() # Subtract word
    SUBU      = enum.auto() # SUBtract Unsigned word
    AND       = enum.auto() # AND
    OR        = enum.auto() # OR
    XOR       = enum.auto() # eXclusive OR
    NOR       = enum.auto() # Not OR

    SLT       = enum.auto() # Set on Less Than
    SLTU      = enum.auto() # Set on Less Than Unsigned
    DADD      = enum.auto() # Doubleword Add
    DADDU     = enum.auto() # Doubleword Add Unsigned
    DSUB      = enum.auto() # Doubleword SUBtract
    DSUBU     = enum.auto() # Doubleword SUBtract Unsigned

    TGE       = enum.auto() # Trap if Greater or Equal
    TGEU      = enum.auto() # Trap if Greater or Equal Unsigned
    TLT       = enum.auto() # Trap if Less Than
    TLTU      = enum.auto() # Trap if Less Than Unsigned
    TEQ       = enum.auto() # Trap if EQual

    TNE       = enum.auto() # Trap if Not Equal

    DSLL      = enum.auto() # Doubleword Shift Left Logical

    DSRL      = enum.auto() # Doubleword Shift Right Logical
    DSRA      = enum.auto() # Doubleword Shift Right Arithmetic
    DSLL32    = enum.auto() # Doubleword Shift Left Logical plus 32

    DSRL32    = enum.auto() # Doubleword Shift Right Logical plus 32
    DSRA32    = enum.auto() # Doubleword Shift Right Arithmetic plus 32

    BLTZ      = enum.auto() # Branch on Less Than Zero
    BGEZ      = enum.auto() # Branch on Greater than or Equal to Zero
    BLTZL     = enum.auto() # Branch on Less Than Zero Likely
    BGEZL     = enum.auto() # Branch on Greater than or Equal to Zero Likely

    TGEI      = enum.auto()
    TGEIU     = enum.auto()
    TLTI      = enum.auto()
    TLTIU     = enum.auto()

    BLTZAL    = enum.auto()
    BGEZAL    = enum.auto()
    BLTZALL   = enum.auto()
    BGEZALL   = enum.auto()

    TEQI      = enum.auto()
    TNEI      = enum.auto()

    J         = enum.auto() # Jump
    JAL       = enum.auto() # Jump And Link
    BEQ       = enum.auto() # Branch on EQual
    BNE       = enum.auto() # Branch on Not Equal
    BLEZ      = enum.auto() # Branch on Less than or Equal to Zero
    BGTZ      = enum.auto() # Branch on Greater Than Zero

    ADDI      = enum.auto() # Add Immediate
    ADDIU     = enum.auto() # Add Immediate Unsigned Word
    SLTI      = enum.auto() # Set on Less Than Immediate
    SLTIU     = enum.auto() # Set on Less Than Immediate Unsigned
    ANDI      = enum.auto() # And Immediate
    ORI       = enum.auto() # Or Immediate
    XORI      = enum.auto() # eXclusive OR Immediate
    LUI       = enum.auto() # Load Upper Immediate

    MFC0      = enum.auto() # Move word From CP0
    DMFC0     = enum.auto() # Doubleword Move From CP0
    CFC0      = enum.auto() # Move control word From CP0

    MTC0      = enum.auto() # Move word to CP0
    DMTC0     = enum.auto() # Doubleword Move To CP0
    CTC0      = enum.auto() # Move control word To CP0

    TLBR      = enum.auto() # Read Indexed TLB Entry
    TLBWI     = enum.auto() # Write Indexed TLB Entry
    TLBWR     = enum.auto() # Write Random TLB Entry
    TLBP      = enum.auto() # Probe TLB for Matching Entry
    ERET      = enum.auto() # Return from Exception

    BC0T      = enum.auto() # Branch on FP True
    BC0F      = enum.auto() # Branch on FP False
    BC0TL     = enum.auto() # Branch on FP True Likely
    BC0FL     = enum.auto() # Branch on FP False Likely

    MFC1      = enum.auto() # Move Word From Floating-Point
    DMFC1     = enum.auto() # Doubleword Move From Floating-Point
    CFC1      = enum.auto() # Move Control Word from Floating-Point

    MTC1      = enum.auto() # Move Word to Floating-Point
    DMTC1     = enum.auto() # Doubleword Move To Floating-Point
    CTC1      = enum.auto() # Move Control Word to Floating-Point

    BC1F      = enum.auto()
    BC1T      = enum.auto()
    BC1FL     = enum.auto()
    BC1TL     = enum.auto()
    ADD_S     = enum.auto() # Floating-Point Add
    SUB_S     = enum.auto() # Floating-Point Sub
    MUL_S     = enum.auto() # Floating-Point Multiply
    DIV_S     = enum.auto() # Floating-Point Divide
    SQRT_S    = enum.auto() # Floating-Point Square Root
    ABS_S     = enum.auto() # Floating-Point Absolute Value
    MOV_S     = enum.auto() # Floating-Point Move
    NEG_S     = enum.auto() # Floating-Point Negate
    ROUND_L_S = enum.auto() # Floating-Point Round to Long Fixed-Point
    TRUNC_L_S = enum.auto() # Floating-Point Truncate to Long Fixed-Point
    CEIL_L_S  = enum.auto() # Floating-Point Ceiling Convert to Long Fixed-Point
    FLOOR_L_S = enum.auto() # Floating-Point Floor Convert to Long Fixed-Point
    ROUND_W_S = enum.auto() # Floating-Point Round to Word Fixed-Point
    TRUNC_W_S = enum.auto() # Floating-Point Truncate to Word Fixed-Point
    CEIL_W_S  = enum.auto() # Floating-Point Ceiling Convert to Word Fixed-Point
    FLOOR_W_S = enum.auto() # Floating-Point Floor Convert to Word Fixed-Point
    CVT_D_S   = enum.auto()
    CVT_W_S   = enum.auto()
    CVT_L_S   = enum.auto()
    C_F_S     = enum.auto()
    C_UN_S    = enum.auto()
    C_EQ_S    = enum.auto()
    C_UEQ_S   = enum.auto()
    C_OLT_S   = enum.auto()
    C_ULT_S   = enum.auto()
    C_OLE_S   = enum.auto()
    C_ULE_S   = enum.auto()
    C_SF_S    = enum.auto()
    C_NGLE_S  = enum.auto()
    C_SEQ_S   = enum.auto()
    C_NGL_S   = enum.auto()
    C_LT_S    = enum.auto()
    C_NGE_S   = enum.auto()
    C_LE_S    = enum.auto()
    C_NGT_S   = enum.auto()
    ADD_D     = enum.auto() # Floating-Point Add
    SUB_D     = enum.auto() # Floating-Point Sub
    MUL_D     = enum.auto() # Floating-Point Multiply
    DIV_D     = enum.auto() # Floating-Point Divide
    SQRT_D    = enum.auto() # Floating-Point Square Root
    ABS_D     = enum.auto() # Floating-Point Absolute Value
    MOV_D     = enum.auto() # Floating-Point Move
    NEG_D     = enum.auto() # Floating-Point Negate
    ROUND_L_D = enum.auto() # Floating-Point Round to Long Fixed-Point
    TRUNC_L_D = enum.auto() # Floating-Point Truncate to Long Fixed-Point
    CEIL_L_D  = enum.auto() # Floating-Point Ceiling Convert to Long Fixed-Point
    FLOOR_L_D = enum.auto() # Floating-Point Floor Convert to Long Fixed-Point
    ROUND_W_D = enum.auto() # Floating-Point Round to Word Fixed-Point
    TRUNC_W_D = enum.auto() # Floating-Point Truncate to Word Fixed-Point
    CEIL_W_D  = enum.auto() # Floating-Point Ceiling Convert to Word Fixed-Point
    FLOOR_W_D = enum.auto() # Floating-Point Floor Convert to Word Fixed-Point
    CVT_S_D   = enum.auto()
    CVT_W_D   = enum.auto()
    CVT_L_D   = enum.auto()
    C_F_D     = enum.auto()
    C_UN_D    = enum.auto()
    C_EQ_D    = enum.auto()
    C_UEQ_D   = enum.auto()
    C_OLT_D   = enum.auto()
    C_ULT_D   = enum.auto()
    C_OLE_D   = enum.auto()
    C_ULE_D   = enum.auto()
    C_SF_D    = enum.auto()
    C_NGLE_D  = enum.auto()
    C_SEQ_D   = enum.auto()
    C_NGL_D   = enum.auto()
    C_LT_D    = enum.auto()
    C_NGE_D   = enum.auto()
    C_LE_D    = enum.auto()
    C_NGT_D   = enum.auto()
    CVT_S_W   = enum.auto()
    CVT_D_W   = enum.auto()
    CVT_S_L   = enum.auto()
    CVT_D_L   = enum.auto()

    BEQL      = enum.auto() # Branch on EQual Likely
    BNEL      = enum.auto() # Branch on Not Equal Likely
    BLEZL     = enum.auto() # Branch on Less than or Equal to Zero Likely
    BGTZL     = enum.auto() # Branch on Greater Than Zero Likely

    DADDI     = enum.auto() # Doubleword add Immediate
    DADDIU    = enum.auto() # Doubleword add Immediate Unsigned
    LDL       = enum.auto() # Load Doubleword Left
    LDR       = enum.auto() # Load Doubleword Right

    LB        = enum.auto() # Load Byte
    LH        = enum.auto() # Load Halfword
    LWL       = enum.auto() # Load Word Left
    LW        = enum.auto() # Load Word
    LBU       = enum.auto() # Load Byte Insigned
    LHU       = enum.auto() # Load Halfword Unsigned
    LWR       = enum.auto() # Load Word Right
    LWU       = enum.auto() # Load Word Unsigned

    SB        = enum.auto() # Store Byte
    SH        = enum.auto() # Store Halfword
    SWL       = enum.auto() # Store Word Left
    SW        = enum.auto() # Store Word
    SDL       = enum.auto() # Store Doubleword Left
    SDR       = enum.auto() # Store Doubleword Right
    SWR       = enum.auto() # Store Word Right
    CACHE     = enum.auto() # Cache

    LL        = enum.auto() # Load Linked word
    LWC1      = enum.auto() # Load Word to Coprocessor z
    LWC2      = enum.auto() # Load Word to Coprocessor z
    PREF      = enum.auto() # Prefetch
    LLD       = enum.auto() # Load Linked Doubleword
    LDC1      = enum.auto() # Load Doubleword to Coprocessor z
    LDC2      = enum.auto() # Load Doubleword to Coprocessor z
    LD        = enum.auto() # Load Doubleword

    SC        = enum.auto() # Store Conditional word
    SWC1      = enum.auto() # Store Word from Coprocessor z
    SWC2      = enum.auto() # Store Word from Coprocessor z
    #
    SCD       = enum.auto() # Store Conditional Doubleword
    SDC1      = enum.auto() # Store Doubleword from Coprocessor z
    SDC2      = enum.auto() # Store Doubleword from Coprocessor z
    SD        = enum.auto() # Store Doubleword

    # Pseudo-Instruction Unique IDs
    BEQZ      = enum.auto() # Branch on EQual Zero
    BNEZ      = enum.auto() # Branch on Not Equal Zero
    B         = enum.auto() # Branch (unconditional)
    NOP       = enum.auto() # No OPeration
    MOVE      = enum.auto() # Move
    NEGU      = enum.auto() 
    NOT       = enum.auto() # Not


@enum.unique
class InstructionVectorId(enum.Enum):
    INVALID   = -1

    VMULF     = enum.auto()
    VMULU     = enum.auto()
    VRNDP     = enum.auto()
    VMULQ     = enum.auto()
    VMUDL     = enum.auto()
    VMUDM     = enum.auto()
    VMUDN     = enum.auto()
    VMUDH     = enum.auto()
    VMACF     = enum.auto()
    VMACU     = enum.auto()
    VRNDN     = enum.auto()
    VMACQ     = enum.auto()
    VMADL     = enum.auto()
    VMADM     = enum.auto()
    VMADN     = enum.auto()
    VMADH     = enum.auto()
    VADD      = enum.auto()
    VSUB      = enum.auto()
    VABS      = enum.auto()
    VADDC     = enum.auto()
    VSUBC     = enum.auto()
    VSAR      = enum.auto()
    VAND      = enum.auto()
    VNAND     = enum.auto()
    VOR       = enum.auto()
    VNOR      = enum.auto()
    VXOR      = enum.auto()
    VNXOR     = enum.auto()

    VLT       = enum.auto()
    VEQ       = enum.auto()
    VNE       = enum.auto()
    VGE       = enum.auto()
    VCL       = enum.auto()
    VCH       = enum.auto()
    VCR       = enum.auto()
    VMRG      = enum.auto()

    VRCP      = enum.auto()
    VRCPL     = enum.auto()
    VRCPH     = enum.auto()
    VMOV      = enum.auto()
    VRSQ      = enum.auto()
    VRSQL     = enum.auto()
    VRSQH     = enum.auto()
    VNOP      = enum.auto()

    MFC2      = enum.auto()
    MTC2      = enum.auto()
    CFC2      = enum.auto()
    CTC2      = enum.auto()

    SBV       = enum.auto()
    SSV       = enum.auto()
    SLV       = enum.auto()
    SDV       = enum.auto()

    SQV       = enum.auto()
    SRV       = enum.auto()

    SPV       = enum.auto()

    SUV       = enum.auto()
    SWV       = enum.auto()

    SHV       = enum.auto()

    SFV       = enum.auto()
    STV       = enum.auto()

    LBV       = enum.auto()
    LSV       = enum.auto()
    LLV       = enum.auto()
    LDV       = enum.auto()

    LQV       = enum.auto()
    LRV       = enum.auto()

    LPV       = enum.auto()

    LUV       = enum.auto()

    LHV       = enum.auto()

    LFV       = enum.auto()
    LTV       = enum.auto()

@enum.unique
class InstrType(enum.Enum):
    typeUnknown = -1

    typeJ       = enum.auto()
    typeI       = enum.auto()
    typeR       = enum.auto()

    typeRegimm  = enum.auto()

@dataclasses.dataclass
class InstrDescriptor:
    operands: list[str]

    instrType: InstrType

    isBranch: bool = False
    isBranchLikely: bool = False
    isJump: bool = False
    isTrap: bool = False

    isFloat: bool = False
    isDouble: bool = False

    isUnsigned: bool = False

    modifiesRt: bool = False
    modifiesRd: bool = False

    mipsVersion: int|None = None
    "Version in which this instruction was introduced. `None` means unknown"
    isRsp: bool = False


instructionDescriptorDict: dict[InstructionId|InstructionVectorId, InstrDescriptor] = {
    InstructionId.INVALID   : InstrDescriptor(["rs", "rt", "IMM"], InstrType.typeUnknown),

    # OP rs
    InstructionId.JR        : InstrDescriptor(["rs"], InstrType.typeR, isJump=True),
    InstructionId.MTHI      : InstrDescriptor(["rs"], InstrType.typeR),
    InstructionId.MTLO      : InstrDescriptor(["rs"], InstrType.typeR),
    InstructionId.JALR      : InstrDescriptor(["rs"], InstrType.typeR, isJump=True, modifiesRd=True),

    # OP rd, rs
    InstructionId.JALR_RD   : InstrDescriptor(["rd", "rs"], InstrType.typeR, isJump=True, modifiesRd=True),

    # OP rd
    InstructionId.MFHI      : InstrDescriptor(["rd"], InstrType.typeR, modifiesRd=True),
    InstructionId.MFLO      : InstrDescriptor(["rd"], InstrType.typeR, modifiesRd=True),

    # OP rs, rt
    InstructionId.MULT      : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.MULTU     : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.DMULT     : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.DMULTU    : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.TGE       : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),
    InstructionId.TGEU      : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),
    InstructionId.TLT       : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),
    InstructionId.TLTU      : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),
    InstructionId.TEQ       : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),
    InstructionId.TNE       : InstrDescriptor(["rs", "rt"], InstrType.typeR, isTrap=True),

    # OP rd, rs, rt
    InstructionId.MOVZ      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.MOVN      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.DIV       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR),
    InstructionId.DIVU      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR),
    InstructionId.SN64_DIV  : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.SN64_DIVU : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.DDIV      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR),
    InstructionId.DDIVU     : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR),
    # InstructionId.DDIV      : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    # InstructionId.DDIVU     : InstrDescriptor(["rs", "rt"], InstrType.typeR),
    InstructionId.ADD       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.ADDU      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.SUB       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.SUBU      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.AND       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.OR        : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.XOR       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.NOR       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.SLT       : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.SLTU      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.DADD      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.DADDU     : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSUB      : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSUBU     : InstrDescriptor(["rd", "rs", "rt"], InstrType.typeR, modifiesRd=True),

    # OP code
    InstructionId.SYSCALL   : InstrDescriptor(["code"], InstrType.typeR),
    InstructionId.BREAK     : InstrDescriptor(["code"], InstrType.typeR),
    InstructionId.SYNC      : InstrDescriptor(["code"], InstrType.typeR),

    # OP rd, rt, rs
    InstructionId.DSLLV     : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRLV     : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRAV     : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.SLLV      : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.SRLV      : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.SRAV      : InstrDescriptor(["rd", "rt", "rs"], InstrType.typeR, modifiesRd=True),

    # OP rd, rt, sa
    InstructionId.SLL       : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.SRL       : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.SRA       : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSLL      : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRL      : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRA      : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSLL32    : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRL32    : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),
    InstructionId.DSRA32    : InstrDescriptor(["rd", "rt", "sa"], InstrType.typeR, modifiesRd=True),

    # OP rs, IMM
    InstructionId.BLTZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BGEZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BLTZL     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.BGEZL     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.TGEI      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),
    InstructionId.TGEIU     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),
    InstructionId.TLTI      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),
    InstructionId.TLTIU     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),
    InstructionId.BLTZAL    : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BGEZAL    : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BLTZALL   : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.BGEZALL   : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.TEQI      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),
    InstructionId.TNEI      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isTrap=True),

    # OP LABEL
    InstructionId.J         : InstrDescriptor(["LABEL"], InstrType.typeJ, isJump=True),
    InstructionId.JAL       : InstrDescriptor(["LABEL"], InstrType.typeJ, isJump=True),

    # OP rs, rt, IMM
    InstructionId.BEQ       : InstrDescriptor(["rs", "rt", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BNE       : InstrDescriptor(["rs", "rt", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BEQL      : InstrDescriptor(["rs", "rt", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.BNEL      : InstrDescriptor(["rs", "rt", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),

    # OP rs, IMM
    InstructionId.BLEZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BGTZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BLEZL     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),
    InstructionId.BGTZL     : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True, isBranchLikely=True),

    # OP rt, IMM
    InstructionId.LUI       : InstrDescriptor(["rt", "IMM"], InstrType.typeI, isUnsigned=True, modifiesRt=True),

    # OP rt, rs, IMM
    InstructionId.ANDI      : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, isUnsigned=True, modifiesRt=True),
    InstructionId.ORI       : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, isUnsigned=True, modifiesRt=True),
    InstructionId.XORI      : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, isUnsigned=True, modifiesRt=True),
    InstructionId.ADDI      : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),
    InstructionId.ADDIU     : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),
    InstructionId.DADDI     : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),
    InstructionId.DADDIU    : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),
    InstructionId.SLTI      : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),
    InstructionId.SLTIU     : InstrDescriptor(["rt", "rs", "IMM"], InstrType.typeI, modifiesRt=True),

    # OP rt, IMM(base)
    InstructionId.LDL       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LDR       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LB        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LH        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LWL       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LW        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LBU       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LHU       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LWR       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LWU       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.SB        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SH        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SWL       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SW        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SDL       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SDR       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.SWR       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI),
    InstructionId.LL        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.PREF      : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LLD       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.LD        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.SC        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.SCD       : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),
    InstructionId.SD        : InstrDescriptor(["rt", "IMM(base)"], InstrType.typeI, modifiesRt=True),

    InstructionId.CACHE     : InstrDescriptor(["op", "IMM(base)"], InstrType.typeI),

    # OP ft, IMM(base)
    InstructionId.LWC1      : InstrDescriptor(["ft", "IMM(base)"], InstrType.typeI, isFloat=True),
    InstructionId.LDC1      : InstrDescriptor(["ft", "IMM(base)"], InstrType.typeI, isFloat=True, isDouble=True),
    InstructionId.SWC1      : InstrDescriptor(["ft", "IMM(base)"], InstrType.typeI, isFloat=True),
    InstructionId.SDC1      : InstrDescriptor(["ft", "IMM(base)"], InstrType.typeI, isFloat=True, isDouble=True),

    # OP cop2t, IMM(base)
    InstructionId.LWC2      : InstrDescriptor(["cop2t", "IMM(base)"], InstrType.typeI),
    InstructionId.LDC2      : InstrDescriptor(["cop2t", "IMM(base)"], InstrType.typeI),
    InstructionId.SWC2      : InstrDescriptor(["cop2t", "IMM(base)"], InstrType.typeI),
    InstructionId.SDC2      : InstrDescriptor(["cop2t", "IMM(base)"], InstrType.typeI),

    # OP rt, cop0d
    InstructionId.MFC0      : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown, modifiesRt=True),
    InstructionId.DMFC0     : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown, modifiesRt=True),
    InstructionId.CFC0      : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown, modifiesRt=True),
    InstructionId.MTC0      : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown),
    InstructionId.DMTC0     : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown),
    InstructionId.CTC0      : InstrDescriptor(["rt", "cop0d"], InstrType.typeUnknown),

    # OP
    InstructionId.TLBR      : InstrDescriptor([], InstrType.typeUnknown),
    InstructionId.TLBWI     : InstrDescriptor([], InstrType.typeUnknown),
    InstructionId.TLBWR     : InstrDescriptor([], InstrType.typeUnknown),
    InstructionId.TLBP      : InstrDescriptor([], InstrType.typeUnknown),
    InstructionId.ERET      : InstrDescriptor([], InstrType.typeUnknown),

    # OP IMM
    InstructionId.BC0T      : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isFloat=True),
    InstructionId.BC0F      : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isFloat=True),
    InstructionId.BC0TL     : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isBranchLikely=True, isFloat=True),
    InstructionId.BC0FL     : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isBranchLikely=True, isFloat=True),

    # OP rt, fs
    InstructionId.MFC1      : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True, modifiesRt=True),
    InstructionId.DMFC1     : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True, modifiesRt=True),
    InstructionId.CFC1      : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True, modifiesRt=True),
    InstructionId.MTC1      : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.DMTC1     : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CTC1      : InstrDescriptor(["rt", "fs"], InstrType.typeUnknown, isFloat=True),

    # OP IMM
    InstructionId.BC1F      : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isFloat=True),
    InstructionId.BC1T      : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isFloat=True),
    InstructionId.BC1FL     : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isBranchLikely=True, isFloat=True),
    InstructionId.BC1TL     : InstrDescriptor(["IMM"], InstrType.typeUnknown, isBranch=True, isBranchLikely=True, isFloat=True),

    # OP fd, fs, ft
    InstructionId.ADD_S     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.SUB_S     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.MUL_S     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.DIV_S     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.ADD_D     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.SUB_D     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.MUL_D     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.DIV_D     : InstrDescriptor(["fd", "fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),

    # OP fd, fs
    InstructionId.SQRT_S    : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.ABS_S     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.MOV_S     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.NEG_S     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.SQRT_D    : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.ABS_D     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.MOV_D     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.NEG_D     : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.ROUND_L_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.TRUNC_L_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CEIL_L_S  : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.FLOOR_L_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.ROUND_L_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.TRUNC_L_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CEIL_L_D  : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.FLOOR_L_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.ROUND_W_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.TRUNC_W_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CEIL_W_S  : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.FLOOR_W_S : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.ROUND_W_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.TRUNC_W_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CEIL_W_D  : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.FLOOR_W_D : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),

    # OP fs, ft
    InstructionId.C_F_S     : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_UN_S    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_EQ_S    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_UEQ_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_OLT_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_ULT_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_OLE_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_ULE_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_F_D     : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_UN_D    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_EQ_D    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_UEQ_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_OLT_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_ULT_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_OLE_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_ULE_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_SF_S    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_NGLE_S  : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_SEQ_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_NGL_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_LT_S    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_NGE_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_LE_S    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_NGT_S   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True),
    InstructionId.C_SF_D    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_NGLE_D  : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_SEQ_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_NGL_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_LT_D    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_NGE_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_LE_D    : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.C_NGT_D   : InstrDescriptor(["fs", "ft"], InstrType.typeUnknown, isFloat=True, isDouble=True),

    # OP fd, fs
    InstructionId.CVT_S_D   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CVT_S_W   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CVT_S_L   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CVT_D_S   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CVT_D_W   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CVT_D_L   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CVT_W_S   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CVT_W_D   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),
    InstructionId.CVT_L_S   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True),
    InstructionId.CVT_L_D   : InstrDescriptor(["fd", "fs"], InstrType.typeUnknown, isFloat=True, isDouble=True),

    # Pseudo-Instruction Unique IDs
    # OP rs, IMM
    InstructionId.BEQZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),
    InstructionId.BNEZ      : InstrDescriptor(["rs", "IMM"], InstrType.typeRegimm, isBranch=True),

    # OP IMM
    InstructionId.B         : InstrDescriptor(["IMM"], InstrType.typeRegimm, isBranch=True),

    # OP
    InstructionId.NOP       : InstrDescriptor([], InstrType.typeR),

    # OP rd, rs
    InstructionId.MOVE      : InstrDescriptor(["rd", "rs"], InstrType.typeR, modifiesRd=True),
    InstructionId.NOT       : InstrDescriptor(["rd", "rs"], InstrType.typeR, modifiesRd=True),

    # OP rd, rt
    InstructionId.NEGU      : InstrDescriptor(["rd", "rt"], InstrType.typeR, modifiesRd=True),
}


InstructionsNotEmitedByIDO = {
    InstructionId.ADD,
    InstructionId.ADDI,
    InstructionId.MTC0,
    InstructionId.MFC0,
    InstructionId.ERET,
    InstructionId.TLBP,
    InstructionId.TLBR,
    InstructionId.TLBWI,
    InstructionId.CACHE,
}



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionCoprocessor0
from .MipsInstructionConfig import InstructionConfig


class InstructionCoprocessor0Rsp(InstructionCoprocessor0):
    Cop0Opcodes_ByFormat = {
        0b00_000: InstructionId.MFC0,
        0b00_100: InstructionId.MTC0,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        self.uniqueId = self.Cop0Opcodes_ByFormat.get(self.fmt, InstructionId.INVALID)


    def disassembleInstruction(self, immOverride: str|None=None) -> str:
        opcode = self.getOpcodeName()
        formated_opcode = opcode.lower().ljust(InstructionConfig.OPCODE_LJUST + self.extraLjustWidthOpcode, ' ')
        # rt = self.getRegisterName(self.rt)
        # rd = self.getCop0RegisterName(self.rd)
        rt = self.getGprRspRegisterName(self.rt)
        rd = self.getCop0RspRegisterName(self.rd)

        if not self.isImplemented():
            return super().disassembleInstruction(immOverride)

        result = f"{formated_opcode} {rt},"
        result = result.ljust(14, ' ')
        result += f" {rd}"
        return result



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import pathlib

import spimdisasm


def exampleMain():
    description = "Single file disassembler example"
    parser = argparse.ArgumentParser(prog="spimdisasm", description=description)

    parser.add_argument("binary", help="Path to input binary")

    parser.add_argument("--output", help="Path to output. Use '-' to print to stdout instead. Defaults to '-'", default="-")

    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {spimdisasm.__version__}")

    parser_singleFile = parser.add_argument_group("Single file disassembly options")

    parser_singleFile.add_argument("--start", help="Raw offset of the input binary file to start disassembling. Expects an hex value", default="0")
    parser_singleFile.add_argument("--end", help="Offset end of the input binary file to start disassembling. Expects an hex value",  default="0xFFFFFF")
    parser_singleFile.add_argument("--vram", help="Set the VRAM address. Expects an hex value")

    args = parser.parse_args()

    # Context is used to store information that should be shared between file sections, such as mapping the symbol vram's to its name and more
    context = spimdisasm.common.Context()

    # Read whole binary input file
    array_of_bytes = spimdisasm.common.Utils.readFileAsBytearray(args.binary)
    inputName = pathlib.Path(args.binary).stem

    start = int(args.start, 16)
    end = int(args.end, 16)

    fileVram = None
    if args.vram is not None:
        fileVram = int(args.vram, 16)

    # Truncate binary to the requested range
    truncatedInputBytes = array_of_bytes[start:end]

    # Asume the input is a .text section. Insntance a SectionText and analyze it
    textSection = spimdisasm.mips.sections.SectionText(context, fileVram, inputName, truncatedInputBytes)
    textSection.analyze()
    textSection.setCommentOffset(start)

    # Write the processed section to a file. This method handles '-' to stdout too
    textSection.saveToFile(args.output)


if __name__ == "__main__":
    exampleMain()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionBase


class InstructionCoprocessor2(InstructionBase):

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionVectorId, InstructionCoprocessor2
from .MipsInstructionConfig import InstructionConfig


class InstructionCoprocessor2Rsp(InstructionCoprocessor2):
    Cop2Opcodes: dict[int, InstructionVectorId] = {
        0x00: InstructionVectorId.VMULF,
        0x01: InstructionVectorId.VMULU,
        0b000_010: InstructionVectorId.VRNDP,
        0b000_011: InstructionVectorId.VMULQ,
        0x04: InstructionVectorId.VMUDL,
        0x05: InstructionVectorId.VMUDM,
        0x06: InstructionVectorId.VMUDN,
        0x07: InstructionVectorId.VMUDH,
        0x08: InstructionVectorId.VMACF,
        0x09: InstructionVectorId.VMACU,
        0b001_010: InstructionVectorId.VRNDN,
        0b001_011: InstructionVectorId.VMACQ,
        0x0C: InstructionVectorId.VMADL,
        0x0D: InstructionVectorId.VMADM,
        0x0E: InstructionVectorId.VMADN,
        0x0F: InstructionVectorId.VMADH,
        0x10: InstructionVectorId.VADD,
        0b010_001: InstructionVectorId.VSUB,
        0b010_011: InstructionVectorId.VABS,
        0x14: InstructionVectorId.VADDC,
        0b010_101: InstructionVectorId.VSUBC,
        0x1D: InstructionVectorId.VSAR,
        0x28: InstructionVectorId.VAND,
        0x29: InstructionVectorId.VNAND,
        0x2A: InstructionVectorId.VOR,
        0x2B: InstructionVectorId.VNOR,
        0x2C: InstructionVectorId.VXOR,
        0x2D: InstructionVectorId.VNXOR,

        0x20: InstructionVectorId.VLT,
        0x21: InstructionVectorId.VEQ,
        0x22: InstructionVectorId.VNE,
        0x23: InstructionVectorId.VGE,
        0x24: InstructionVectorId.VCL,
        0x25: InstructionVectorId.VCH,
        0x26: InstructionVectorId.VCR,
        0x27: InstructionVectorId.VMRG,

        0b110_000: InstructionVectorId.VRCP,
        0b110_001: InstructionVectorId.VRCPL,
        0b110_010: InstructionVectorId.VRCPH,
        0b110_011: InstructionVectorId.VMOV,
        0b110_100: InstructionVectorId.VRSQ,
        0b110_101: InstructionVectorId.VRSQL,
        0b110_110: InstructionVectorId.VRSQH,
        0b110_111: InstructionVectorId.VNOP,
    }
    Cop2MoveOpcodes: dict[int, InstructionVectorId] = {
        0b00_000: InstructionVectorId.MFC2,
        0b00_100: InstructionVectorId.MTC2,
        0b00_010: InstructionVectorId.CFC2,
        0b00_110: InstructionVectorId.CTC2,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        super().processUniqueId()

        self.uniqueId = self.Cop2Opcodes.get(self.function, InstructionVectorId.INVALID)
        if self[25] == 0:
            self.uniqueId = self.Cop2MoveOpcodes.get(self.elementHigh, InstructionVectorId.INVALID)


    def modifiesRt(self) -> bool:
        if self.uniqueId in (InstructionVectorId.CFC2, InstructionVectorId.MFC2):
            return True
        return super().modifiesRt()


    def disassembleInstruction(self, immOverride: str|None=None) -> str:
        opcode = self.getOpcodeName()
        formated_opcode = opcode.lower().ljust(InstructionConfig.OPCODE_LJUST + self.extraLjustWidthOpcode, ' ')
        e_upper = self[25]
        e = self.processVectorElement(self.elementHigh)
        vt = self.getVectorRspRegisterName(self.vt)
        vs = self.getVectorRspRegisterName(self.vs)
        vd = self.getVectorRspRegisterName(self.vd)

        result = formated_opcode

        if self.uniqueId == InstructionVectorId.VNOP:
            return opcode.lower()

        if self.uniqueId in (InstructionVectorId.VMOV, InstructionVectorId.VRCP, InstructionVectorId.VRCPH, InstructionVectorId.VRSQ, InstructionVectorId.VRSQH, InstructionVectorId.VRSQL):
            result += f" {vd}[{self.vs}], "
            result = result.ljust(14, ' ')
            result += f" {vt}[{e}]"
            return result

        if e_upper == 0:
            rt = f"${self.rt}"
            rd = f"${self.rd}"
            result += f" {rt},"
            result = result.ljust(14, ' ')
            if self.uniqueId in (InstructionVectorId.CFC2, InstructionVectorId.CTC2):
                result += f" {rd}"
            else:
                # TODO: improve
                index = self.sa>>1
                # TODO: use vector register instead of rd
                result += f" {rd}[{index}]"
        else:
            result += f" {vd},"
            result = result.ljust(14, ' ')
            result += f" {vs},"
            result = result.ljust(19, ' ')
            result += f" {vt}"
            if self.elementHigh != 0:
                # TODO: do this properly
                result += f"[{e}]"

        if not self.isImplemented():
            result = "ERROR # " + result
        return result



#!/usr/bin/env python3

from setuptools import find_packages, setup # type: ignore

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="spimdisasm",
    version="1.0.5",
    author="Decompollaborate",
    license="MIT",
    description="N64 MIPS disassembler",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

# Relocation format used by overlays of Zelda64, Yoshi Story and Doubutsu no Mori (Animal Forest)

from __future__ import annotations

from ... import common

from .. import symbols

from ..MipsRelocTypes import RelocTypes

from . import SectionBase


class RelocEntry:
    def __init__(self, entry: int):
        self.sectionId = entry >> 30
        self.relocType = (entry >> 24) & 0x3F
        self.offset = entry & 0x00FFFFFF

    @property
    def reloc(self):
        return (self.sectionId << 30) | (self.relocType << 24) | (self.offset)

    def getSectionType(self) -> common.FileSectionType:
        return common.FileSectionType.fromId(self.sectionId)

    def getRelocType(self) -> RelocTypes:
        return RelocTypes.fromValue(self.relocType)

    def __str__(self) -> str:
        section = self.getSectionType().toStr()
        reloc = self.getRelocType().name
        return f"{section} {reloc} 0x{self.offset:X}"
    def __repr__(self) -> str:
        return self.__str__()


class SectionRelocZ64(SectionBase):
    def __init__(self, context: common.Context, vram: int|None, filename: str, array_of_bytes: bytearray):
        super().__init__(context, vram, filename, array_of_bytes, common.FileSectionType.Reloc)

        self.seekup = self.words[-1]

        self.setCommentOffset(self.sizew*4 - self.seekup)

        # Remove non reloc stuff
        self.words = self.words[-self.seekup // 4:]

        self.sectionSizes = {
            common.FileSectionType.Text: self.words[0],
            common.FileSectionType.Data: self.words[1],
            common.FileSectionType.Rodata: self.words[2],
            common.FileSectionType.Bss: self.words[3],
        }
        self.relocCount = self.words[4]

        self.tail = self.words[self.relocCount+5:-1]

        self.entries: list[RelocEntry] = list()
        for word in self.words[5:self.relocCount+5]:
            self.entries.append(RelocEntry(word))

        self.differentSegment: bool = False

    @property
    def nRelocs(self) -> int:
        return len(self.entries)

    @property
    def textSize(self) -> int:
        return self.sectionSizes[common.FileSectionType.Text]
    @property
    def dataSize(self) -> int:
        return self.sectionSizes[common.FileSectionType.Data]
    @property
    def rodataSize(self) -> int:
        return self.sectionSizes[common.FileSectionType.Rodata]
    @property
    def bssSize(self) -> int:
        return self.sectionSizes[common.FileSectionType.Bss]


    def analyze(self):
        localOffset = 0

        currentVram = self.getVramOffset(localOffset)
        sym = symbols.SymbolData(self.context, localOffset + self.inFileOffset, currentVram, f"{self.name}_OverlayInfo", self.words[0:4])
        sym.setCommentOffset(self.commentOffset)
        sym.endOfLineComment = [f" # _{self.name}Segment{sectName.toCapitalizedStr()}Size" for sectName in common.FileSections_ListBasic]
        sym.analyze()
        self.symbolList.append(sym)
        localOffset += 4 * 4

        currentVram = self.getVramOffset(localOffset)
        sym = symbols.SymbolData(self.context, localOffset + self.inFileOffset, currentVram, f"{self.name}_RelocCount", [self.relocCount])
        sym.setCommentOffset(self.commentOffset)
        sym.analyze()
        self.symbolList.append(sym)
        localOffset += 4

        currentVram = self.getVramOffset(localOffset)
        sym = symbols.SymbolData(self.context, localOffset + self.inFileOffset, currentVram, f"{self.name}_OverlayRelocations", [r.reloc for r in self.entries])
        sym.setCommentOffset(self.commentOffset)
        sym.endOfLineComment = [f" # {str(r)}" for r in self.entries]
        sym.analyze()
        self.symbolList.append(sym)
        localOffset += 4 * len(self.entries)

        if len(self.tail) > 0:
            currentVram = self.getVramOffset(localOffset)
            sym = symbols.SymbolData(self.context, localOffset + self.inFileOffset, currentVram, f"{self.name}_Padding", self.tail)
            sym.setCommentOffset(self.commentOffset)
            sym.analyze()
            self.symbolList.append(sym)
            localOffset += 4 * len(self.tail)

        currentVram = self.getVramOffset(localOffset)
        sym = symbols.SymbolData(self.context, localOffset + self.inFileOffset, currentVram, f"{self.name}_OverlayInfoOffset", [self.seekup])
        sym.setCommentOffset(self.commentOffset)
        sym.analyze()
        self.symbolList.append(sym)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import enum


class AbiNames(enum.Enum):
    numeric = enum.auto()
    o32 = enum.auto()
    n32 = enum.auto()
    n64 = enum.auto()

    @staticmethod
    def fromStr(x: str) -> AbiNames:
        if x in ("32", "o32"):
            return AbiNames.o32
        if x in ("n32",):
            return AbiNames.n32
        if x in ("64", "n64"):
            return AbiNames.n64
        return AbiNames.numeric


class InstructionConfig:
    NAMED_REGISTERS: bool = True
    """Enables using named registers

    This option takes precedence over the other named register options"""

    GPR_ABI_NAMES: AbiNames = AbiNames.o32
    """The ABI names to be used for general purpose registers when disassembling the main processor's instructions"""
    FPR_ABI_NAMES: AbiNames = AbiNames.numeric
    """The ABI names to be used for floating point registers when disassembling the floating point (coprocessor 1) instructions"""

    USE_FPCCSR: bool = True
    """Use FpcCsr as register $31 for the FP control/status register"""

    VR4300_COP0_NAMED_REGISTERS: bool = True
    """Use named registers for VR4300's coprocessor 0 registers"""
    VR4300_RSP_COP0_NAMED_REGISTERS: bool = True
    """Use named registers for VR4300's RSP's coprocessor 0 registers"""

    PSEUDO_INSTRUCTIONS: bool = True
    """Produce pseudo instructions (like `move`, `nop` or `b`) whenever those should match the desired original instruction"""

    PSEUDO_BEQZ: bool = True
    PSEUDO_BNEZ: bool = True
    PSEUDO_B: bool = True
    PSEUDO_MOVE: bool = True
    PSEUDO_NOT: bool = True
    PSEUDO_NEGU: bool = True


    SN64_DIV_FIX: bool = False
    """Enables a few fixes for SN64's assembler related to div/divu instructions

    - SN64's assembler doesn't like assembling `div $0, a, b` with .set noat active.
    Removing the $0 fixes this issue (but not for handwritten asm)

    - SN64's assembler expands div to have break if dividing by zero
    However, the break it generates is different than the one it generates with `break N`
    So we replace break instrutions for SN64 with the exact word that the assembler generates when expanding div
    """

    OPCODE_LJUST: int = 7+4
    """The minimal number of characters to left-align the opcode name"""

    UNKNOWN_INSTR_COMMENT: bool = True
    """Generate a pseudo-disassembly comment when disassembling non implemented instructions"""

    @staticmethod
    def addParametersToArgParse(parser: argparse.ArgumentParser):
        registerNames = parser.add_argument_group("MIPS register names options")

        registerNames.add_argument("--named-registers", help=f"(Dis)allows named registers for every instruction. This flag takes precedence over similar flags in this category. Defaults to {InstructionConfig.NAMED_REGISTERS}", action=argparse.BooleanOptionalAction)

        abi_choices = ["numeric", "32", "o32", "n32", "n64"]
        registerNames.add_argument("--Mgpr-names", help=f"Use GPR names according to the specified ABI. Defaults to {InstructionConfig.GPR_ABI_NAMES.name}", choices=abi_choices)
        registerNames.add_argument("--Mfpr-names", help=f"Use FPR names according to the specified ABI. Defaults to {InstructionConfig.FPR_ABI_NAMES.name}", choices=abi_choices)
        registerNames.add_argument("--Mreg-names", help=f"Use GPR and FPR names according to the specified ABI. This flag takes precedence over --Mgpr-names and --Mfpr-names", choices=abi_choices)

        registerNames.add_argument("--use-fpccsr", help=f"Toggles using the FpcCsr alias for float register $31 when using the numeric ABI. Defaults to {InstructionConfig.USE_FPCCSR}", action=argparse.BooleanOptionalAction)

        registerNames.add_argument("--cop0-named-registers", help=f"Toggles using the built-in names for registers of the VR4300's Coprocessor 0. Defaults to {InstructionConfig.USE_FPCCSR}", action=argparse.BooleanOptionalAction)
        registerNames.add_argument("--rsp-cop0-named-registers", help=f"Toggles using the built-in names for registers of the RSP's Coprocessor 0. Defaults to {InstructionConfig.USE_FPCCSR}", action=argparse.BooleanOptionalAction)


        miscOpts = parser.add_argument_group("MIPS misc instructions options")

        miscOpts.add_argument("--pseudo-instr", help=f"Toggles producing pseudo instructions. Defaults to {InstructionConfig.PSEUDO_INSTRUCTIONS}", action=argparse.BooleanOptionalAction)

        miscOpts.add_argument("--sn64-div-fix", help=f"Enables a few fixes for SN64's assembler related to div/divu instructions. Defaults to {InstructionConfig.SN64_DIV_FIX}", action=argparse.BooleanOptionalAction)

        miscOpts.add_argument("--opcode-ljust", help=f"Set the minimal number of characters to left-align the opcode name. Defaults to {InstructionConfig.OPCODE_LJUST}")

        miscOpts.add_argument("--unk-instr-comment", help=f"Disables the extra comment produced after unknown instructions. Defaults to {InstructionConfig.UNKNOWN_INSTR_COMMENT}", action=argparse.BooleanOptionalAction)


    @classmethod
    def parseArgs(cls, args: argparse.Namespace):
        if args.named_registers is not None:
            InstructionConfig.NAMED_REGISTERS = args.named_registers

        if args.Mgpr_names:
            InstructionConfig.GPR_ABI_NAMES = AbiNames.fromStr(args.Mgpr_names)
        if args.Mfpr_names:
            InstructionConfig.FPR_ABI_NAMES = AbiNames.fromStr(args.Mfpr_names)
        if args.Mreg_names:
            InstructionConfig.GPR_ABI_NAMES = AbiNames.fromStr(args.Mreg_names)
            InstructionConfig.FPR_ABI_NAMES = AbiNames.fromStr(args.Mreg_names)

        if args.use_fpccsr is not None:
            InstructionConfig.USE_FPCCSR = args.use_fpccsr

        if args.cop0_named_registers is not None:
            InstructionConfig.VR4300_COP0_NAMED_REGISTERS = args.cop0_named_registers
        if args.rsp_cop0_named_registers is not None:
            InstructionConfig.VR4300_RSP_COP0_NAMED_REGISTERS = args.rsp_cop0_named_registers

        if args.pseudo_instr is not None:
            InstructionConfig.PSEUDO_INSTRUCTIONS = args.pseudo_instr

        if args.sn64_div_fix is not None:
            InstructionConfig.SN64_DIV_FIX = args.sn64_div_fix

        if args.opcode_ljust is not None:
            InstructionConfig.OPCODE_LJUST = int(args.opcode_ljust, 0)

        if args.unk_instr_comment is not None:
            InstructionConfig.UNKNOWN_INSTR_COMMENT = args.unk_instr_comment



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from .. import symbols

from . import SectionBase


class SectionRodata(SectionBase):
    def __init__(self, context: common.Context, vram: int|None, filename: str, array_of_bytes: bytearray):
        super().__init__(context, vram, filename, array_of_bytes, common.FileSectionType.Rodata)

        self.bytes: bytearray = bytearray(self.sizew*4)
        common.Utils.beWordsToBytes(self.words, self.bytes)

        # addresses of symbols in this rodata section
        self.symbolsVRams: set[int] = set()


    def _stringGuesser(self, contextSym: common.ContextSymbolBase, localOffset: int) -> bool:
        if contextSym.isMaybeString or contextSym.isString():
            return True

        if not common.GlobalConfig.STRING_GUESSER:
            return False

        if not contextSym.hasNoType() or contextSym.referenceCounter > 1:
            return False

        # This would mean the string is an empty string, which is not very likely
        if self.bytes[localOffset] == 0:
            return False

        try:
            common.Utils.decodeString(self.bytes, localOffset)
        except (UnicodeDecodeError, RuntimeError):
            # String can't be decoded
            return False
        return True

    def _processElfRelocSymbols(self) -> None:
        if len(self.context.relocSymbols[self.sectionType]) == 0:
            return

        # Process reloc symbols (probably from a .elf file)
        inFileOffset = self.inFileOffset
        for w in self.words:
            relocSymbol = self.context.getRelocSymbol(inFileOffset, self.sectionType)
            if relocSymbol is not None:
                if relocSymbol.name.startswith("."):
                    sectType = common.FileSectionType.fromStr(relocSymbol.name)
                    relocSymbol.sectionType = sectType

                    relocName = f"{relocSymbol.name}_{w:06X}"
                    contextOffsetSym = common.ContextOffsetSymbol(w, relocName, sectType)
                    if sectType == common.FileSectionType.Text:
                        # jumptable
                        relocName = f"L{w:06X}"
                        contextOffsetSym = self.context.addOffsetJumpTableLabel(w, relocName, common.FileSectionType.Text)
                        relocSymbol.type = contextOffsetSym.type
                        offsetSym = self.context.getOffsetSymbol(inFileOffset, self.sectionType)
                        if offsetSym is not None:
                            offsetSym.type = common.SymbolSpecialType.jumptable
                    self.context.offsetSymbols[sectType][w] = contextOffsetSym
                    relocSymbol.name = relocName
                    # print(relocSymbol.name, f"{w:X}")
            inFileOffset += 4

    def analyze(self):
        self.checkAndCreateFirstSymbol()

        symbolList = []
        localOffset = 0

        partOfJumpTable = False
        for w in self.words:
            currentVram = self.getVramOffset(localOffset)
            contextSym = self.context.getAnySymbol(currentVram)

            if contextSym is not None and contextSym.isJumpTable():
                partOfJumpTable = True

            elif partOfJumpTable:
                if localOffset in self.pointersOffsets:
                    partOfJumpTable = True

                elif self.context.getGenericSymbol(currentVram) is not None:
                    partOfJumpTable = False

                elif ((w >> 24) & 0xFF) != 0x80:
                    partOfJumpTable = False

            if partOfJumpTable:
                labelSym = self.context.addJumpTableLabel(w, isAutogenerated=True)
                labelSym.referenceCounter += 1

            elif currentVram in self.context.newPointersInData:
                if common.GlobalConfig.ADD_NEW_SYMBOLS:
                    contextSym = self.context.getAnySymbol(currentVram)
                    if contextSym is None:
                        contextSym = self.context.addSymbol(currentVram, None, self.sectionType, isAutogenerated=True)
                        contextSym.isDefined = True

                    contextSym.isMaybeString = self._stringGuesser(contextSym, localOffset)
                    self.context.newPointersInData.remove(currentVram)

            elif contextSym is not None:
                contextSym.isMaybeString = self._stringGuesser(contextSym, localOffset)

            contextSym = self.context.getAnySymbol(currentVram)
            if contextSym is not None:
                self.symbolsVRams.add(currentVram)
                contextSym.isDefined = True

                symbolList.append((localOffset, currentVram, contextSym.name))

            localOffset += 4

        for i, (offset, vram, symName) in enumerate(symbolList):
            if i + 1 == len(symbolList):
                words = self.words[offset//4:]
            else:
                nextOffset = symbolList[i+1][0]
                words = self.words[offset//4:nextOffset//4]

            symVram = None
            if self.vram is not None:
                symVram = vram

            sym = symbols.SymbolRodata(self.context, offset + self.inFileOffset, symVram, symName, words)
            sym.setCommentOffset(self.commentOffset)
            sym.analyze()
            self.symbolList.append(sym)

        self._processElfRelocSymbols()


    def removePointers(self) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = super().removePointers()
        for i in range(self.sizew):
            top_byte = (self.words[i] >> 24) & 0xFF
            if top_byte == 0x80:
                self.words[i] = top_byte << 24
                was_updated = True
            if (top_byte & 0xF0) == 0x00 and (top_byte & 0x0F) != 0x00:
                self.words[i] = top_byte << 24
                was_updated = True

        return was_updated



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import dataclasses
import struct


# a.k.a. Ehdr (elf header)
@dataclasses.dataclass
class Elf32Header:
    ident:      list[int]  # 16 bytes  # 0x00
    type:       int        # half      # 0x10
    machine:    int        # half      # 0x12
    version:    int        # word      # 0x14
    entry:      int        # address   # 0x18
    phoff:      int        # offset    # 0x1C
    shoff:      int        # offset    # 0x20
    flags:      int        # word      # 0x24
    ehsize:     int        # half      # 0x28
    phentsize:  int        # half      # 0x2A
    phnum:      int        # half      # 0x2C
    shentsize:  int        # half      # 0x2E
    shnum:      int        # half      # 0x30
    shstrndx:   int        # half      # 0x32
                                       # 0x34

    @staticmethod
    def fromBytearray(array_of_bytes: bytearray, offset: int = 0) -> Elf32Header:
        identFormat = ">16B"
        ident = list(struct.unpack_from(identFormat, array_of_bytes, 0 + offset))
        # print(ident)

        headerFormat = ">HHIIIIIHHHHHH"
        unpacked = struct.unpack_from(headerFormat, array_of_bytes, 0x10 + offset)
        # print(unpacked)

        return Elf32Header(ident, *unpacked)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common


class SymbolBase(common.ElementBase):
    def __init__(self, context: common.Context, inFileOffset: int, vram: int|None, name: str, words: list[int], sectionType: common.FileSectionType):
        super().__init__(context, inFileOffset, vram, name, words, sectionType)

        self.endOfLineComment: list[str] = []

        self.contextSym: common.ContextSymbol|None = None
        if self.vram is not None:
            contextSym = self.context.getAnySymbol(self.vram)
            if contextSym is None:
                symName: str|None = self.name
                if symName == "":
                    symName = None
                contextSym = self.context.addSymbol(self.vram, symName, self.sectionType)
                contextSym.isAutogenerated = True
                contextSym.isDefined = True
            self.contextSym = contextSym
            self.name = contextSym.name


    def generateAsmLineComment(self, localOffset: int, wordValue: int|None = None) -> str:
        if not common.GlobalConfig.ASM_COMMENT:
            return ""

        offsetHex = f"{localOffset + self.inFileOffset + self.commentOffset:06X}"

        vramHex = ""
        if self.vram is not None:
            currentVram = self.getVramOffset(localOffset)
            vramHex = f"{currentVram:08X} "

        wordValueHex = ""
        if wordValue is not None:
            wordValueHex = f"{common.Utils.beWordToCurrenEndian(wordValue):08X} "

        return f"/* {offsetHex} {vramHex}{wordValueHex}*/"

    def getSymbolAtVramOrOffset(self, localOffset: int) -> common.ContextSymbolBase | None:
        if self.vram is not None:
            currentVram = self.getVramOffset(localOffset)
            return self.context.getAnySymbol(currentVram)
        return self.context.getOffsetSymbol(self.inFileOffset + localOffset, self.sectionType)

    def getLabel(self) -> str:
        if self.contextSym is not None:
            return self.getLabelFromSymbol(self.contextSym)

        offsetSym = self.context.getOffsetSymbol(self.inFileOffset, self.sectionType)
        return self.getLabelFromSymbol(offsetSym)


    def isRdata(self) -> bool:
        "Checks if the current symbol is .rdata"
        return False


    def renameBasedOnSection(self):
        if not common.GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE:
            return

        if self.sectionType != common.FileSectionType.Rodata and self.sectionType != common.FileSectionType.Bss:
            return

        if self.vram is None:
            return

        if self.contextSym is not None:
            if not self.contextSym.isAutogenerated:
                return

            if self.sectionType == common.FileSectionType.Rodata:
                if self.contextSym.type != common.SymbolSpecialType.jumptable:
                    self.contextSym.name = f"R_{self.vram:08X}"
            elif self.sectionType == common.FileSectionType.Bss:
                self.contextSym.name = f"B_{self.vram:08X}"
            self.name = self.contextSym.name

    def renameBasedOnType(self):
        pass


    def analyze(self):
        self.renameBasedOnSection()
        self.renameBasedOnType()

        if self.contextSym is not None:
            byteStep = 4
            if self.contextSym.isByte():
                byteStep = 1
            elif self.contextSym.isShort():
                byteStep = 2

            for i in range(0, self.sizew):
                localOffset = 4*i
                for j in range(0, 4, byteStep):
                    if i == 0 and j == 0:
                        continue
                    contextSym = self.getSymbolAtVramOrOffset(localOffset+j)
                    if contextSym is not None:
                        contextSym.isDefined = True
                        contextSym.sectionType = self.sectionType
                        if contextSym.hasNoType():
                            contextSym.type = contextSym.type


    def getNthWord(self, i: int, canReferenceSymbolsWithAddends: bool=False, canReferenceConstants: bool=False) -> tuple[str, int]:
        output = ""
        localOffset = 4*i
        w = self.words[i]

        isByte = False
        isShort = False
        if self.contextSym is not None:
            if self.contextSym.isByte():
                isByte = True
            elif self.contextSym.isShort():
                isShort = True

        dotType = ".word"
        byteStep = 4
        if isByte:
            dotType = ".byte"
            byteStep = 1
        elif isShort:
            dotType = ".short"
            byteStep = 2

        for j in range(0, 4, byteStep):
            label = ""
            if j != 0 or i != 0:
                contextSym = self.getSymbolAtVramOrOffset(localOffset+j)
                if contextSym is not None:
                    # Possible symbols in the middle
                    label = common.GlobalConfig.LINE_ENDS + contextSym.getSymbolLabel()  + common.GlobalConfig.LINE_ENDS

            if isByte:
                shiftValue = 24 - (j * 8)
                subVal = (w & (0xFF << shiftValue)) >> shiftValue
                value = f"0x{subVal:02X}"
            elif isShort:
                shiftValue = 16 - (j * 8)
                subVal = (w & (0xFFFF << shiftValue)) >> shiftValue
                value = f"0x{subVal:04X}"
            else:
                value = f"0x{w:08X}"

                # .elf relocated symbol
                if len(self.context.relocSymbols[self.sectionType]) > 0:
                    possibleReference = self.context.getRelocSymbol(self.inFileOffset + localOffset, self.sectionType)
                    if possibleReference is not None:
                        value = possibleReference.getNamePlusOffset(w)
                else:
                    # This word could be a reference to a symbol
                    symbolRef = self.context.getGenericSymbol(w, tryPlusOffset=canReferenceSymbolsWithAddends)
                    if symbolRef is not None:
                        value = symbolRef.getSymbolPlusOffset(w)
                    elif canReferenceConstants:
                        constant = self.context.getConstant(w)
                        if constant is not None:
                            value = constant.getName()

            comment = self.generateAsmLineComment(localOffset+j)
            output += f"{label}{comment} {dotType} {value}"
            if j == 0 and i < len(self.endOfLineComment):
                output += self.endOfLineComment[i]
            output += common.GlobalConfig.LINE_ENDS

        return output, 0


    def disassembleAsData(self) -> str:
        output = self.getLabel()

        canReferenceSymbolsWithAddends = self.vram in self.context.dataSymbolsWithReferencesWithAddends
        canReferenceConstants = self.vram in self.context.dataReferencingConstants

        i = 0
        while i < self.sizew:
            data, skip = self.getNthWord(i, canReferenceSymbolsWithAddends, canReferenceConstants)
            output += data

            i += skip
            i += 1
        return output

    def disassemble(self) -> str:
        return self.disassembleAsData()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionRegimm


class InstructionRegimmRsp(InstructionRegimm):
    RemovedOpcodes: dict[int, InstructionId] = {
        0b00_010: InstructionId.BLTZL,
        0b00_011: InstructionId.BGEZL,

        0b01_000: InstructionId.TGEI,
        0b01_001: InstructionId.TGEIU,
        0b01_010: InstructionId.TLTI,
        0b01_011: InstructionId.TLTIU,

        0b10_010: InstructionId.BLTZALL,
        0b10_011: InstructionId.BGEZALL,

        0b01_100: InstructionId.TEQI,
        0b01_110: InstructionId.TNEI,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        if self.rt not in self.RemovedOpcodes:
            self.uniqueId = self.RegimmOpcodes.get(self.rt, InstructionId.INVALID)


    def getRegisterName(self, register: int) -> str:
        return self.getGprRspRegisterName(register)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionBase, instructionDescriptorDict
from .MipsInstructionConfig import InstructionConfig


class InstructionCoprocessor0(InstructionBase):
    Cop0Opcodes_ByFormat = {
        0b00_000: InstructionId.MFC0,
        0b00_001: InstructionId.DMFC0,
        0b00_010: InstructionId.CFC0,
        # 0b00_011: "",
        0b00_100: InstructionId.MTC0,
        0b00_101: InstructionId.DMTC0,
        0b00_110: InstructionId.CTC0,
        # 0b00_111: "",
    }
    Cop0Opcodes_ByFunction = {
        0b000_001: InstructionId.TLBR,
        0b000_010: InstructionId.TLBWI,
        0b000_110: InstructionId.TLBWR,
        0b001_000: InstructionId.TLBP,
        0b011_000: InstructionId.ERET,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        if self.fmt in InstructionCoprocessor0.Cop0Opcodes_ByFormat:
            self.uniqueId = InstructionCoprocessor0.Cop0Opcodes_ByFormat[self.fmt]
        elif self.fmt == 0b01_000: # fmt = BC
            if self.tf:
                if self.nd:
                    self.uniqueId = InstructionId.BC0TL
                else:
                    self.uniqueId = InstructionId.BC0T
            else:
                if self.nd:
                    self.uniqueId = InstructionId.BC0FL
                else:
                    self.uniqueId = InstructionId.BC0F
        elif self.function in InstructionCoprocessor0.Cop0Opcodes_ByFunction:
            self.uniqueId = InstructionCoprocessor0.Cop0Opcodes_ByFunction[self.function]

        self.descriptor = instructionDescriptorDict[self.uniqueId]

    def blankOut(self):
        if self.fmt in InstructionCoprocessor0.Cop0Opcodes_ByFormat:
            self.rt = 0
            self.rd = 0
            self.sa = 0
            self.function = 0
        elif self.fmt == 0b01_000:
            self.rd = 0
            self.sa = 0
            self.function = 0
        elif self.function in InstructionCoprocessor0.Cop0Opcodes_ByFunction:
            self.rt = 0
            self.rd = 0
            self.sa = 0



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

__version__ = "1.0.5"
__author__ = "Decompollaborate"

from . import common
from . import elf32
from . import mips



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from .. import instructions
from .. import symbols

from ..MipsFileBase import FileBase

from . import SectionBase


class SectionText(SectionBase):
    def __init__(self, context: common.Context, vram: int|None, filename: str, array_of_bytes: bytearray):
        super().__init__(context, vram, filename, array_of_bytes, common.FileSectionType.Text)

        # TODO: do something with this information
        self.fileBoundaries: list[int] = list()

    @property
    def nFuncs(self) -> int:
        return len(self.symbolList)

    @staticmethod
    def wordListToInstructions(wordList: list[int], currentVram: int|None, isRsp: bool=False) -> list[instructions.InstructionBase]:
        instrsList: list[instructions.InstructionBase] = list()
        for word in wordList:
            if isRsp:
                instr = instructions.wordToInstructionRsp(word)
            else:
                instr = instructions.wordToInstruction(word)

            if currentVram is not None:
                instr.vram = currentVram
                currentVram += 4

            instrsList.append(instr)
        return instrsList

    def analyze(self):
        functionEnded = False
        farthestBranch = 0
        funcsStartsList = [0]
        unimplementedInstructionsFuncList = []

        instrsList = self.wordListToInstructions(self.words, self.getVramOffset(0) if self.vram is not None else None, self.isRsp)

        instructionOffset = 0
        currentInstructionStart = 0
        currentFunctionSym = self.context.getFunction(self.getVramOffset(instructionOffset))

        isLikelyHandwritten = self.isHandwritten

        isInstrImplemented = True
        index = 0
        nInstr = len(instrsList)
        while index < nInstr:
            instr = instrsList[index]
            if not instr.isImplemented():
                isInstrImplemented = False

            if functionEnded:
                functionEnded = False

                isLikelyHandwritten = self.isHandwritten
                index += 1
                instructionOffset += 4
                isboundary = False
                # Loop over until we find a instruction that isn't a nop
                while index < nInstr:
                    instr = instrsList[index]
                    if instr.uniqueId != instructions.InstructionId.NOP:
                        if isboundary:
                            self.fileBoundaries.append(self.inFileOffset + index*4)
                        break
                    index += 1
                    instructionOffset += 4
                    isboundary = True

                currentInstructionStart = instructionOffset
                currentFunctionSym = self.context.getFunction(self.getVramOffset(instructionOffset))

                funcsStartsList.append(index)
                unimplementedInstructionsFuncList.append(not isInstrImplemented)
                if index >= len(instrsList):
                    break
                instr = instrsList[index]
                isInstrImplemented = instr.isImplemented()

            currentVram = self.getVramOffset(instructionOffset)

            if not self.isRsp and not isLikelyHandwritten:
                isLikelyHandwritten = instr.isLikelyHandwritten()

            if instr.isBranch() or (common.GlobalConfig.TREAT_J_AS_UNCONDITIONAL_BRANCH and instr.uniqueId == instructions.InstructionId.J):
                if instr.uniqueId == instructions.InstructionId.J:
                    branch = instr.getInstrIndexAsVram() - currentVram
                else:
                    branch = instr.getBranchOffset()
                if branch > farthestBranch:
                    # keep track of the farthest branch target
                    farthestBranch = branch
                if branch < 0:
                    if branch + instructionOffset < 0:
                        # Whatever we are reading is not a valid instruction
                        break
                    # make sure to not branch outside of the current function
                    if not isLikelyHandwritten:
                        j = len(funcsStartsList) - 1
                        while j >= 0:
                            if (branch + instructionOffset) < funcsStartsList[j] * 4:
                                vram = self.getVramOffset(funcsStartsList[j]*4)
                                funcSymbol = self.context.getFunction(vram)
                                if funcSymbol is not None and funcSymbol.isTrustableFunction(self.isRsp):
                                    j -= 1
                                    continue
                                del funcsStartsList[j]
                                del unimplementedInstructionsFuncList[j-1]
                            else:
                                break
                            j -= 1

            elif instr.isJType():
                target = instr.getInstrIndexAsVram()
                if not self.isRsp:
                    if target >= 0x84000000:
                        # RSP address space?
                        isLikelyHandwritten = True
                funcSym = self.context.getFunction(target)
                if funcSym is None:
                    if instr.uniqueId == instructions.InstructionId.J and not self.isRsp:
                        funcSym = self.context.addFakeFunction(target, f".L{target:08X}", isAutogenerated=True)
                    else:
                        funcSym = self.context.addFunction(target, None, isAutogenerated=True)

            if not (farthestBranch > 0):
                if instr.uniqueId == instructions.InstructionId.JR:
                    if instr.rs == 31: # $ra
                        functionEnded = True
                elif instr.uniqueId == instructions.InstructionId.J:
                    if isLikelyHandwritten or self.isRsp:
                        functionEnded = True

            vram = currentVram + 8
            funcSymbol = self.context.getFunction(vram)
            if funcSymbol is not None and funcSymbol.isTrustableFunction(self.isRsp):
                functionEnded = True

            if currentFunctionSym is not None and currentFunctionSym.size > 4:
                if instructionOffset + 8 == currentInstructionStart + currentFunctionSym.size:
                        functionEnded = True

            index += 1
            farthestBranch -= 4
            instructionOffset += 4

        unimplementedInstructionsFuncList.append(not isInstrImplemented)

        i = 0
        startsCount = len(funcsStartsList)
        for startIndex in range(startsCount):
            start = funcsStartsList[startIndex]
            hasUnimplementedIntrs = unimplementedInstructionsFuncList[startIndex]
            end = nInstr
            if startIndex + 1 < startsCount:
                end = funcsStartsList[startIndex+1]

            if start >= end:
                break

            funcName = f"func_{i}"
            if len(self.context.offsetSymbols[self.sectionType]) > 0:
                possibleFuncName = self.context.getOffsetSymbol(start*4, self.sectionType)
                if possibleFuncName is not None:
                    funcName = possibleFuncName.name

            vram = None
            if self.vram is not None:
                vram = self.getVramOffset(start*4)

                if common.GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS or not hasUnimplementedIntrs:
                    funcSymbol = self.context.getFunction(vram)
                    if funcSymbol is None:
                        funcSymbol = self.context.addFunction(vram, None, isAutogenerated=True)
                    funcSymbol.isDefined = True
                else:
                    if vram in self.context.symbols:
                        self.context.symbols[vram].isDefined = True
                    elif common.GlobalConfig.ADD_NEW_SYMBOLS:
                        contextSym = self.context.addSymbol(vram, None)
                        contextSym.isAutogenerated = True
                        contextSym.isDefined = True

            func = symbols.SymbolFunction(self.context, self.inFileOffset + start*4, vram, funcName, instrsList[start:end])
            func.index = i
            func.pointersOffsets |= self.pointersOffsets
            func.hasUnimplementedIntrs = hasUnimplementedIntrs
            func.parent = self
            func.isRsp = self.isRsp
            func.analyze()
            self.symbolList.append(func)
            i += 1

    def printAnalyzisResults(self):
        super().printAnalyzisResults()
        if not common.GlobalConfig.PRINT_NEW_FILE_BOUNDARIES:
            return

        nBoundaries = len(self.fileBoundaries)
        if nBoundaries > 0:
            print(f"File {self.name}")
            print(f"Found {self.nFuncs} functions.")
            print(f"Found {nBoundaries} file boundaries.")

            print("\t offset, size, vram\t functions")

            for i in range(len(self.fileBoundaries)-1):
                start = self.fileBoundaries[i]
                end = self.fileBoundaries[i+1]

                functionsInBoundary = 0
                for func in self.symbolList:
                    funcOffset = func.inFileOffset - self.inFileOffset
                    if start <= funcOffset < end:
                        functionsInBoundary += 1
                fileVram = 0
                if self.vram is not None:
                    fileVram = start + self.vram
                print("\t", f"{start+self.commentOffset:06X}", f"{end-start:04X}", f"{fileVram:08X}", "\t functions:", functionsInBoundary)

            start = self.fileBoundaries[-1]
            end = self.sizew*4 + self.inFileOffset

            functionsInBoundary = 0
            for func in self.symbolList:
                if func.vram is not None and self.vram is not None:
                    funcOffset = func.vram - self.vram
                    if start <= funcOffset < end:
                        functionsInBoundary += 1
            fileVram = 0
            if self.vram is not None:
                fileVram = start + self.vram
            print("\t", f"{start+self.commentOffset:06X}", f"{end-start:04X}", f"{fileVram:08X}", "\t functions:", functionsInBoundary)

            print()
        return


    def compareToFile(self, other: FileBase):
        result = super().compareToFile(other)

        if isinstance(other, SectionText):
            result["text"] = {
                "diff_opcode": self.countDiffOpcodes(other),
                "same_opcode_same_args": self.countSameOpcodeButDifferentArguments(other),
            }

        return result

    def countDiffOpcodes(self, other: SectionText) -> int:
        result = 0
        for i in range(min(self.nFuncs, other.nFuncs)):
            func = self.symbolList[i]
            other_func = other.symbolList[i]
            assert isinstance(func, symbols.SymbolFunction)
            assert isinstance(other_func, symbols.SymbolFunction)
            result += func.countDiffOpcodes(other_func)
        return result

    def countSameOpcodeButDifferentArguments(self, other: SectionText) -> int:
        result = 0
        for i in range(min(self.nFuncs, other.nFuncs)):
            func = self.symbolList[i]
            other_func = other.symbolList[i]
            assert isinstance(func, symbols.SymbolFunction)
            assert isinstance(other_func, symbols.SymbolFunction)
            result += func.countSameOpcodeButDifferentArguments(other_func)
        return result

    def blankOutDifferences(self, other_file: FileBase) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        if not isinstance(other_file, SectionText):
            return False

        was_updated = False
        for i in range(min(self.nFuncs, other_file.nFuncs)):
            func = self.symbolList[i]
            other_func = other_file.symbolList[i]
            assert isinstance(func, symbols.SymbolFunction)
            assert isinstance(other_func, symbols.SymbolFunction)
            was_updated = func.blankOutDifferences(other_func) or was_updated

        return was_updated

    def removePointers(self) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = False
        for func in self.symbolList:
            assert isinstance(func, symbols.SymbolFunction)
            was_updated = func.removePointers() or was_updated

        return was_updated

    def removeTrailingNops(self) -> bool:
        was_updated = False

        if self.nFuncs > 0:
            func = self.symbolList[-1]
            assert isinstance(func, symbols.SymbolFunction)
            func.removeTrailingNops()
            was_updated = True

        return was_updated
